<'
# Copyright (c) 2012-2015,2017 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Mar  7 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include "<"$comp">_control_task.h"
#include "<"$comp">_activity.h"


/* <"[--- genom_${comp}_activity_alloc ----------------------------------]"> */

int
genom_<"$comp">_activity_alloc(struct genom_activities *activities)
{
  static int aid = 0;
  size_t id;

  /* look for the first free activity */
  for(id = 0; id < genom_max_activities(); id++) {
    if (activities->a[id].any.status == ACT_VOID) {
      activities->a[id].any.aid = ++aid;
      activities->a[id].any.status = ACT_INIT;
      activities->a[id].any.start = 0;
      activities->a[id].any.stop = 0;
      activities->a[id].any.pause = 0;
      return id;
    }
  }

  return -1;
}


/* === Send final replies ================================================== */

<'foreach s [$component services] {'>
/* <"[$s name]"> */
static __inline__ void
genom_<"$comp">_<"[$s name]">_activity_report(
  struct genom_component_data *self,
  struct genom_<"$comp">_<"[$s name]">_activity *a)
{
  STATUS s;

  /* send final reply */
  if (a->h.state == <"$comp">_ether) {
    /* success: update after/before array */
    self->control.run_map[<"$COMP">_<"[$s name]">_RQSTID] = 1;

    s = csServReplySend(
      self->control.csserv, a->h.rid, FINAL_REPLY, OK,
      (char *)&a->out, 0, genom_<"$comp">_<"[$s name]">_encode);
  } else
    s = csServReplySend(
      self->control.csserv, a->h.rid, FINAL_REPLY, ERROR, (char *)&a->h, 0,
      genom_<"$comp">_activity_encodex);
  if (s == ERROR) {
    genom_log_warn(1, "could not send output for service %s", "<"[$s name]">");
    s = csServReplySend(
      self->control.csserv, a->h.rid, FINAL_REPLY, ERROR, NULL, 0,
      genom_<"$comp">_genom_serialization_encodex);
    if (s == ERROR) {
      genom_log_warn(1, "discarding output for service %s", "<"[$s name]">");
    } else
      genom_log_warn(0, "invalid output for service %s", "<"[$s name]">");
  }

  /* cleanup */
  genom_tfini_<"$comp">_<"[$s name]">_activity(a);
}

<'}'>

/* <"[--- genom_${comp}_activity_report ---------------------------------]"> */

void
genom_<"$comp">_activity_report(
  struct genom_component_data *self, struct genom_activity *a)
{
  switch(a->sid) {
    case -1: return; /* permanent activity reports nothing */

<'foreach s [$component services] {'>
    case <"$COMP">_<"[$s name]">_RQSTID:
      genom_<"$comp">_<"[$s name]">_activity_report(
        self, (struct genom_<"$comp">_<"[$s name]">_activity *)a);
      return;
<'}'>
  }

  assert(!"code not reached");
}


/* --- genom_<"$comp">_activity_encodex ------------------------------------ */

int
genom_<"$comp">_activity_encodex(char *buffer, int size, char *dst, int maxsize)
{
  struct genom_activity *a = (struct genom_activity *)buffer;

  if (a->state == genom_unkex_id)
    return genom_<"$comp">_genom_unkex_encodex(a->exdetail, size, dst, maxsize);
  else if (a->state == genom_syserr_id)
    return genom_<"$comp">_genom_syserr_encodex(a->exdetail, size, dst, maxsize);
<'foreach e [$component throws] {'>
  else if (a->state == <"[$e cname]">_id)
    return genom_<"$comp">_<"[$e cname]">_encodex(
      a->exdetail, size, dst, maxsize);
<'}'>

  assert(!"unknown genom exception");
  return 0;
}


/* === Interrupt incompatible activities =================================== */

<'foreach s [$component services] {'>
/* <"[$s name]"> */
static __inline__ int
genom_<"$comp">_<"[$s name]">_interrupt_reqd(
  struct genom_component_data *self, struct genom_activity *a)
{
  int delay = 0;
  (void)self; (void)a; /* fix -Wunused-parameter */

<'  foreach task [$component tasks] {'>
<'    set clist [list]'>
<'    foreach c [$s interrupt] {'>
<'      if {[catch {$c task} t]} continue'>
<'      if {$t eq $task} { lappend clist $c }'>
<'    }'>
<'    if {[llength $clist] == 0} continue'>

  /* task <"[$task name]"> */
  pthread_mutex_lock(&self->tasks.<"[$task name]">.lock);
  {
    struct genom_activity *c;
    size_t id;

    for(id = 0; id < genom_max_activities(); id++) {
      c = &self->tasks.<"[$task name]">.activities.a[id].any;
      if (c->status == ACT_VOID || c->status == ACT_ETHER) continue;
      if (c->aid == a->aid) continue;
      switch(c->sid) {
<'    foreach c $clist {'>
        case <"$COMP">_<"[$c name]">_RQSTID:
<'    }'>
          switch(c->status) {
            case ACT_RUN:
              delay = 1;
              if (!self->tasks.<"[$task name]">.runnable) {
                self->tasks.<"[$task name]">.runnable = 1;
                pthread_cond_broadcast(&self->tasks.<"[$task name]">.sync);
              }
              /*FALLTHROUGH*/
            case ACT_INIT:
              c->stop = 1;
              c->interruptedby = "<"[$s name]">";
              break;

            case ACT_STOP: delay = 1; break;

            default: assert(!"code not reached");
          }
      }
    }
  }
  pthread_mutex_unlock(&self->tasks.<"[$task name]">.lock);
<'  }'>

  return delay;
}

<'}'>

/* <"[--- genom_${comp}_interrupt_reqd ----------------------------------]"> */

int
genom_<"$comp">_interrupt_reqd(
  struct genom_component_data *self, struct genom_activity *a)
{
  /* interrupt incompatible activities */
  switch(a->sid) {
    case -1: return 0; /* permanent activity is not interruptible */

<'foreach s [$component services] {'>
    case <"$COMP">_<"[$s name]">_RQSTID:
      return genom_<"$comp">_<"[$s name]">_interrupt_reqd(self, a);
<'}'>
  }

  assert(!"code not reached");
  return 0;
}
